#ifndef COMM_MIN_LOAD_BALANCE
#define COMM_MIN_LOAD_BALANCE

/*
 * First Order Communication Minimization Second Order Load Balance Algorithm
 * --------------------------------------------------------------------------
 *
 * There are two different algorithms used to run first order communicaition 
 * second order load balance mapping.  The first is algorithm 1 and is implemented in 
 * the DoAlgo1 function the second is algorithm 2 which is implemented in the
 * DoAlgo2 funcion.  Both functions first find the ideal bound of a processor
 * by taking the sum of the utilizations of all of the components and then averaging
 * this amount amoung the processors.  If this average is above any processor's
 * bound the ideal bound of a processor is set to its true bound and the amount
 * of utilization that is left over is then dispersed the rest of the processors.
 * This process is repeated until all of the total utilization is dispersed to 
 * a processor in the form of the processor's ideal bound.  If this can not be 
 * completed then this is an error.
 *
 * Algorithm 1 first sorts all of the nodes in order of the sum of the edge cost
 * for all of its edges.  The edges of the node with the higheset total edge cost
 * are then sorted in order of their edge cost.  The two nodes that are part of the
 * edge with the highest edge cost are then compared to see if they can be merged.
 * 
 * Algorithm 2 first sorts all of the edges in order of their edge cost.  The two
 * nodes that are part of the edge with the highest edge cost are then compared to
 * see if they can be merged.
 *
 * When checking to see if two nodes can be merged both algorithms check to see if
 * these nodes have a processor assigned to them.  If they both do and the processors
 * are different then they cannot be merged.  However, if both processors are the
 * same or only one node has a processor assigned to it the sum of the total utilization
 * of both nodes is checked to see if it is below the ideal bound.  Otherwise each processor
 * is checked to see if then the total utilization of these two nodes and any other tasks
 * on this processor is checked to see if this is below the ideal bound.  If so then these
 * two nodes can be merged.  If the sum of the utilizations is above the ideal bound
 * but lower than the processor's true bound then these nodes are marked as a possible
 * merge.  If there is a previous (best) possible merge the amount of extra utilization
 * above the processor's ideal bound is checked for both the current and best possible
 * merges.  If the current possible merge has a smaller amount of extra utilization then
 * the best possible merge then the current possible merge is marked as the best.
 *
 * If either algorithm goes through all of its connections and there were no two nodes
 * connected by an edge that could be merged without causing a processor to excede its
 * ideal bound the algorithm checks to see if there is a best possible merge.  If so 
 * these two nodes are merged together, if not then the algorithm checks nodes that are
 * not connected.  To try to minimize the extra communication that this will cause
 * the node with the lowest total edge cost is compared with the node with the second
 * lowest total edge cost and so on to see if this creates a valid merge.  This process
 * is the same as checking the connected nodes where if no pair of nodes can be merged
 * to keep a processor below its ideal utilization then the best possible merge is 
 * used.  If no best possible merge is found then the algorithm cannot be completed and
 * no solution was found.
 */

//Run the communication minimization algorithm with second order load balance mapping.
//returns the amount of utilization that could not be mapped to the processor or -1 if
//there is an error but there may be a solution that the algorithm could not find
double Do_CommMinLoadBalance_Mapping();

#endif