#include "comm_min-first_fit.h"

//global includes
#include <map>

//local includes
#include "debug.h"
#include "task_connection.h"
#include "graph.h"
#include "first_fit.h"

using namespace std;

double	abs(double d);	//absolute value for a double

//externals
extern Graph grp;
extern System sys;

//static globals
static CommMinGraph global_graph;

//debug varialbles for comm. min. with second order load balance
bool algo1;
bool algo2;

/*************************************************************************************/
/*************** Communication Minimization with second order First Fit **************/
/*************************************************************************************/
// - communication min. with second order load balance algorithm is below

double Do_CommMinFirstFit_Mapping()
{
	GraphPartition best_partition;
	double non_allocated_utility = 0;

	//reset all processors
	for(list<Processor *>::iterator proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
	{
		Processor* proc_ptr = *proc_it;

		proc_ptr->Clear_Tasks();							//clear all tasks on the processor
		proc_ptr->Property()->Reset_Available_Space();		//give the processor its entire amount of space available
	}

	PopulateOriginalGraph(&global_graph);
	best_partition.SetGroupSize(sys.Get_Number_Of_Procs());	//set the numver of groups to the number of processors
	best_partition.cost = -1;								//set partition to be invalid

	//run the min cut algorithm if all the tasks cannot be mapped to one processor
	//and there are at least 2 processors
	if(!AllTasksOnOneProc(&best_partition) && best_partition.group.size() >= 2)
		MinCutAlgo(&global_graph, &best_partition);

	if(best_partition.cost == -1)							//is the partition valid
	{
		non_allocated_utility = FirstFit();					//try one more time using the first fit algorithm - to get the nonallocated utility 
		if(non_allocated_utility)
			return non_allocated_utility;

		form_partition(&best_partition);					//first fit algorithm worked - form the partition
		best_partition.cost = findPartitionCost(&best_partition, &global_graph);
	}
	else
		map_tasks_to_procs(&best_partition);				//assign the tasks to the processors in the sys variable

	print_CommMin_Mapping(best_partition.cost);				//print out the results

	return 0;
}

void MinCutAlgo(CommMinGraph* graph, GraphPartition* best_partition)
{
//	CommMinGraph rgraph = CommMinGraph(graph);				//copy the graph so original stays intact
	int groups_used = GetNumGroupsUsed(best_partition);
	int groups_avaliable = best_partition->group.size() - groups_used;

	GraphPartition last_partition = *best_partition;

	MinimumCut(best_partition, graph);

	if(best_partition->cost == 0)	//if a partition of 0 is found return - best partition possible
		return;

	//find the best partition
	while(graph->GetNodes().size() >= groups_avaliable && graph->GetNodes().size() > 1)
	{
		MinimumCut(&last_partition, graph);

		//if the last partition is valid and either the best partition so far is invalid
		//or the cost of the las partition is less then the cost of the best partition
		//stor the last partition as the best partition
		if(last_partition.cost != -1 && (best_partition->cost == -1 || last_partition.cost < best_partition->cost))
			*best_partition = last_partition;	

		if(best_partition->cost == 0)
			return;

		//clear out groups that were temperarily used to find the partition
		for(int i = groups_used; i < last_partition.group.size(); ++i)
			last_partition.group[i].clear();
	}

	return;
}

void MinimumCut(GraphPartition* partition, CommMinGraph* graph)
{
	CommMinNode* s = NULL;
	CommMinNode* t = NULL;
	double cost;
	int group_num;

	//find the group that needs to be populated
	for(group_num = 0; group_num < partition->group.size(); ++group_num)
		if(partition->group[group_num].empty())
		{
			group_num++;
			break;
		}

	list<CommMinNode*> nodes = graph->GetNodes();	//copy the nodes so the list can be
													//be modified without changing the
													//actual graph

	s = nodes.front();
	partition->AddTasks(group_num, s);
	nodes.pop_front();

	while(!nodes.empty())
	{
		if(nodes.size() == 1)
		{
			t = nodes.front();
			partition->AddTasks(group_num - 1, t);
			nodes.pop_front();
			break;
		}

		s = FindBestNode(group_num, partition, nodes, graph);	//find the best node to add to the partition
		partition->AddTasks(group_num, s);
		nodes.remove(s);
	}

	if(valid_partition(partition))								//is partition valid
		partition->cost = findPartitionCost(partition, graph);
	else if(partition->group[partition->group.size()-1].size())
		partition->cost = -1;									//the partition does not have any free groups
	else
	{
		CommMinGraph graph_copy = CommMinGraph(graph);

//		list<attachment> attachments = StoreAttachments(t);		//the partition does have free groups
																//split up the graph more	
		CommMinNode* new_t = graph_copy.FindNodeFromTask(t->GetTasks().front());

		//remove the last node added (t) from the copied graph
		graph_copy.RemoveNode(new_t);
		graph_copy.RemoveEdgesFrom(new_t);

		partition->group[group_num].clear();	//clear out the tasks that were in the strongest connected group
		MinCutAlgo(&graph_copy, partition);		//run min cut algorithm on modified graph
	}


	//s and t should always be non NULL - this if statement is just an added precaution
	if(s && t)
	{
		//merge s and t together and remove t and its edges from the graph
		s->Merge(t);
		graph->RemoveNode(t);
		graph->RemoveEdgesFrom(t);

		delete t;
	}

} //MinimumCut

void PopulateOriginalGraph(CommMinGraph* graph)
{
	graph->clean();		//clear out any existing stuff

	//add in all of the nodes - needs to be done before adding edges
	for(list<TaskType*>::iterator task_it = grp.Node_Begin(); task_it != grp.Node_End(); ++task_it)
	{
		TaskType* task_ptr = *task_it;
		Processor* proc;

		if(proc = task_ptr->Property()->Get_Alloc_Proc())
		{
//			double utilization = task_ptr->Property()->Get_Exec_Time() / task_ptr->Property()->Get_Period();
//			if(proc->Property()->Allocate_Space(utilization))
//				proc->Add_Task(task_ptr);
		}

		CommMinNode* new_task = new CommMinNode(task_ptr);
		graph->AddNode(new_task);
	}

	//add in all of the edges
	for(task_it = grp.Node_Begin(); task_it != grp.Node_End(); ++task_it)
	{
		//add in connections that are coming out of this connection
		for(list<ConnectionType*>::iterator conn_it = (*task_it)->Get_Out_Data_Edges_Begin(); 
		conn_it != (*task_it)->Get_Out_Data_Edges_End(); ++conn_it)
		{
			TaskType* from = (*conn_it)->Get_From();
			TaskType* to = (*conn_it)->Get_To();

			//get the node that the tasks are part of 
			CommMinNode* node1 = graph->FindNodeFromTask(from);
			CommMinNode* node2 = graph->FindNodeFromTask(to);

			double bps = (*conn_it)->Property()->Get_Frequency() * (*conn_it)->Property()->Get_Message_Size();

			CommMinEdge* new_edge = new CommMinEdge(bps, node1, node2);

			//add the new edge to its nodes and to the graph
			node1->AddEdge(new_edge);
			node2->AddEdge(new_edge);
			graph->AddEdge(new_edge);
		}
	}
}

CommMinNode* FindBestNode(int group_num, GraphPartition* partition, list<CommMinNode*> nodes, CommMinGraph* graph)
{
	CommMinNode* best;
	double saved_cost;
	double best_saved_cost;

	best			= nodes.front();	//initialize the best node as the first node
	best_saved_cost	= findSavedCost(group_num, best, partition, graph); //find the saved cost of adding it to the partition

	for(list<CommMinNode*>::iterator it = nodes.begin(); it != nodes.end(); ++it)
	{
		//find the saved cost of adding it to the partition
		saved_cost = findSavedCost(group_num, *it, partition, graph);

		//check to see if this node saves more then the previous best node
		if(saved_cost > best_saved_cost)
		{
			best_saved_cost = saved_cost;
			best = *it;
		}
	}

	return best;
}

double findSavedCost(int group_num, CommMinNode* node, GraphPartition* partition, CommMinGraph* graph)
{
	double saved_cost = 0;
	list<TaskType*> tasks = partition->group[group_num];
	list<CommMinNode*> checked_nodes;

	for(list<TaskType*>::iterator it = tasks.begin(); it != tasks.end(); ++it)
	{
		CommMinNode* it_node = graph->FindNodeFromTask(*it);

		//check if this node has already been tested, if so don't check it again
		if(onList(it_node, checked_nodes))
			continue;
		checked_nodes.push_back(it_node);

		list<CommMinEdge*> edges = it_node->GetEdges();

		//go through each of the edges in the partition and see if they are connected 
		//to the given node - if so add in the edge's bps to the saved cost
		for(list<CommMinEdge*>::iterator conn_it = it_node->GetEdgeBegin(); conn_it != it_node->GetEdgeEnd(); ++conn_it)
		{
			CommMinEdge* edge = *conn_it;

			CommMinNode* first_node = edge->GetFirstConnectedNode();
			CommMinNode* second_node = edge->GetSecondConnectedNode();

			if(first_node == node || second_node == node)
				saved_cost += edge->GetMessageSize();
		}
	}

	return saved_cost;
}

double findPartitionCost(GraphPartition* partition, CommMinGraph* graph)
{
	double cost = 0;
	double edge_cost;

	CommMinNode* node1;
	CommMinNode* node2;

	list<CommMinEdge*> edges = global_graph.GetEdges();

	//go through each edge and determine if its nodes are in different groups - if so
	//then add the edge cost to the total cost
	for(list<CommMinEdge*>::iterator edge_it = edges.begin(); edge_it != edges.end(); ++edge_it)
	{
		int node1group;
		int node2group;

		node1		= (*edge_it)->GetFirstConnectedNode();
		node2		= (*edge_it)->GetSecondConnectedNode();
		edge_cost   = (*edge_it)->GetMessageSize();

		node1group  = getGroup(node1, partition, &global_graph);
		node2group  = getGroup(node2, partition, &global_graph);

		if(node1group != node2group)
			cost += edge_cost;
	}

	return cost;
}

bool onList(CommMinNode* node_check, list<CommMinNode*>  checked_nodes)
{
	for(list<CommMinNode*>::iterator it = checked_nodes.begin(); it != checked_nodes.end(); ++it)
		if(node_check == *it) return true;

	return false;
}

int getGroup(CommMinNode* node, GraphPartition* partition, CommMinGraph* graph)
{
	list<CommMinNode*> checked_nodes;

	for(int i = 0; i < partition->group.size(); ++i)
	{
		list<TaskType*> tasks = partition->group[i];
		for(list<TaskType*>::iterator task_it = tasks.begin(); task_it != tasks.end(); ++task_it)
		{
			CommMinNode* node_check = graph->FindNodeFromTask(*task_it);

			//make sure that the node has not already been previously checked
			if(onList(node_check, checked_nodes))
				continue;
			else
				checked_nodes.push_back(node_check);
			
			//found the node
			if(node_check == node)
				return i;
		}
	}

	//could not find the node
	return -1;
}


void show_partition(GraphPartition* partition)
{
	string name;
	int group;

	//go through each group and each task in the group
	for(group = 0; group < partition->group.size(); ++group)
		for(list<TaskType*>::iterator it = partition->group[group].begin(); it != partition->group[group].end(); ++it)
			name = (*it)->Property()->Get_Name();
}

bool valid_partition(GraphPartition* partition)
{
	multiset<list<TaskType*>, partition_group_greater_than> tasks;
	multiset<Processor*, processor_greater_than> procs;

	if(partition->group.size() == 0)	//if there are no groups then the partition cannot be valid
		return false;

	//sort the groups in the partition by their total utilization
	for(int i = 0; i < partition->group.size(); ++i)
		tasks.insert(partition->group[i]);

	//sort the processors by their available space
	for(list<Processor*>::iterator proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
		procs.insert(*proc_it);

	//check to see if the largest group can be mapped to the processor to the greatest
	//bound, the second largest to the second greatest, and so on
	multiset<list<TaskType*>, partition_group_greater_than>::iterator task_it = tasks.begin();
	for(multiset<Processor*, processor_greater_than>::iterator it = procs.begin();
	it != procs.end(); ++it)
	{
		if(GetTotalUtilization(*task_it) > (*it)->Property()->Get_Bound())
			return false;

		MapGroupToProc(*task_it, *it, partition);
		task_it++;
	}

	return true;
}

bool AllTasksOnOneProc(GraphPartition* partition)
{
	list<TaskType*> all_tasks;
	multiset<Processor*, processor_greater_than> procs;

	if(partition->group.size() == 0)	//no procs so cannot put them all on one processor
		return false;

	//get all of the tasks
	for(list<TaskType*>::iterator all_it = grp.Node_Begin(); all_it != grp.Node_End(); ++all_it)
		all_tasks.push_back(*all_it);

	//sort the processors by their available space
	for(list<Processor*>::iterator proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
		procs.insert(*proc_it);

	//can all of the tasks be put on the processory with the largest bound
	if(GetTotalUtilization(all_tasks) < (*procs.begin())->Property()->Get_Bound())
	{
		partition->group[0] = all_tasks;
		partition->procs[0] = *procs.begin();
		partition->cost = 0;
		return true;
	}

	return false;
}

list<attachment> StoreAttachments(CommMinNode* node)
{
	attachment a;
	list<attachment> attachments;

	//go through each edge of the node
	for(list<CommMinEdge*>::iterator edge_it = node->GetEdgeBegin(); edge_it != node->GetEdgeEnd(); ++edge_it)
	{
		double bps = (*edge_it)->GetMessageSize();
		CommMinNode* other_node;

		//find the other node that is part of the edge
		if((*edge_it)->GetFirstConnectedNode() == node)
			other_node = (*edge_it)->GetSecondConnectedNode();
		else
			other_node = (*edge_it)->GetFirstConnectedNode();

		//get all of the tasks that are part of the other node
		for(list<TaskType*>::iterator task_it = other_node->GetTasksBegin(); task_it != other_node->GetTasksEnd(); ++task_it)
		{
			a.bytes_per_second = bps;
			a.task = *task_it;
			attachments.push_back(a);
		}
	}

	return attachments;
}

bool partition_group_greater_than::operator () (list<TaskType*> t1, list<TaskType*> t2)
{
	return GetTotalUtilization(t1) > GetTotalUtilization(t2);
}

double GetTotalUtilization(list<TaskType*> tasks)
{
	double u = 0;

	for(list<TaskType*>::iterator it = tasks.begin(); it != tasks.end(); ++it)
		u += (*it)->Property()->Get_Exec_Time() / (*it)->Property()->Get_Period();

	return u;
}

int GetNumGroupsUsed(GraphPartition* partition)
{
	for(int i = 0; i < partition->group.size(); ++i)
		if(partition->group[i].empty()) break;

	return i;
}

void map_tasks_to_procs(GraphPartition* partition)
{
	for(int i = 0; i < partition->group.size(); ++i)
	{
		for(list<TaskType*>::iterator task_it = partition->group[i].begin(); task_it != partition->group[i].end(); ++task_it)
		{
			double utilization = (*task_it)->Property()->Get_Exec_Time() / (*task_it)->Property()->Get_Period();
			partition->procs[i]->Property()->Allocate_Space(utilization);	//this function should always return true - there is a bug otherwise
			partition->procs[i]->Add_Task(*task_it);
		}
	}
}

void form_partition(GraphPartition* partition)
{
	int i = 0;
	for(list<Processor*>::iterator proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
	{
		for(list<TaskType*>::iterator task_it = (*proc_it)->Tasks_Begin(); task_it != (*proc_it)->Tasks_End(); ++task_it)
			partition->group[i].push_back(*task_it);
		++i;
	}
}

void MapGroupToProc(list<TaskType*> task_list, Processor* proc, GraphPartition* partition)
{
	//find the group that the list represents
	for(int i = 0; i < partition->group.size(); ++i)
		if(task_list.front() == partition->group[i].front()) break;

	//if the group was found set the processor to it
	if(i < partition->group.size())
		partition->procs[i] = proc;
}


/******************************************************************************************/
/*************** Communication Minimization with second order Load Balancing **************/
/******************************************************************************************/

//static helper classes

//struct to keep track of a processor, its ideal_bound (the bound that the processor would
//prefer to stay below but is less then or equal to its true bound), how much utilization
//can be added until the ideal bound is reached, and a node that may be assigned to it 
struct ProcWrapper
{
	ProcWrapper(Processor* in_proc = NULL, double in_bound = 0): proc(in_proc), 
		ideal_bound(in_bound), ideal_space(in_bound), node(NULL) {}

	double ideal_bound;
	double ideal_space;
	Processor* proc;
	CommMinNode* node;
};

//struct to keep track a possible valid merge between two nodes and the extra utilization
//that the merge would cause processors to go over their ideal bound
struct PossibleMerge
{
	double extra_util;
	CommMinNode* node1;
	CommMinNode* node2;
};

//list of tasks and their total utilization
struct TaskList
{
	list<TaskType*> tasks;
	double utilization;
};

//returns the sum of the cost of all of the node's edges 
static double	FindTotalEdgeCost(CommMinNode* node);

/* Function Objecst used for sorting*/
class ProcWrpGreaterThan
{
public:
	bool operator()(ProcWrapper p1, ProcWrapper p2)
	{
		return p1.ideal_space > p2.ideal_space;
	}
};

class TaskListGreaterThan
{
public:
	bool operator()(TaskList list1, TaskList list2)
	{
		return list1.utilization > list2.utilization;
	}
};

class NodeEdgeCostGreaterThan
{
public:
	bool operator()(CommMinNode* node1, CommMinNode* node2)
	{
		return FindTotalEdgeCost(node1) > FindTotalEdgeCost(node2);
	}
};

class EdgeCostGreaterThan
{
public:
	bool operator()(CommMinEdge* edge1, CommMinEdge* edge2)
	{
		return edge1->GetMessageSize() > edge2->GetMessageSize();
	}
};

//static helper functions

//run the first algorithm - Merge the first valid pair of nodes where one node has the greatest total edge 
//cost and the node other node is the one that is connected to the first node with the greatest edge cost
static double	DoAlgo1(multiset<ProcWrapper, ProcWrpGreaterThan>& proc_wrps);

//run the second algorithm - Merge the first valid pair of nodes where the nodes are the two connected by
//the edge with the highest communication cost
static double	DoAlgo2(multiset<ProcWrapper, ProcWrpGreaterThan>& proc_wrps);

//returns the utilization of all of the nodes in the global_graph
static double	FindTotalUtilization();

//populates the given list and determines the ideal bound for each member in the list - returns the amount
//of utilization that could not be put onto a processor, 0 if successful
static double	DetermineProcCommMinBounds(list<ProcWrapper>& wrappers);

//check to see if merging node1 and node2 is valid - if so merge the two nodes, returns true if nodes were
//merged and false if they were not 
static bool		CheckAndMerge(CommMinNode* node1, CommMinNode* node2, multiset<ProcWrapper, ProcWrpGreaterThan>& proc_wrps);

//returns the position of the TaskList in the global variable task_lists
static multiset<TaskList, TaskListGreaterThan>::iterator	GetTaskList(CommMinNode*);

//merge the two nodes and update the global variable task_lists
static void		MergeAndUpdate(CommMinNode* node1, CommMinNode* node2);

//assign processors to the tasks that are in global variable task_lists
static void		AssignProcs(multiset<ProcWrapper, ProcWrpGreaterThan>& proc_wrps);

static bool		MergeUnconnected(PossibleMerge& merged_nodes, multiset<CommMinNode*, NodeEdgeCostGreaterThan> nodes, multimap<CommMinNode*, CommMinNode*> checked, multiset<ProcWrapper, ProcWrpGreaterThan>& proc_wrps);

//debug function to view the tasks in each node
static void		ShowNodes();

//static globals			
static PossibleMerge best_valid_merge;
static multiset<TaskList, TaskListGreaterThan> task_lists;	//list of a list of tasks that keeps its elements in order 
															//by the total utilization of all of the tasks in the list
double Do_CommMinLoadBalance_Mapping()
{
	double rval;
	multiset<ProcWrapper, ProcWrpGreaterThan> ms_proc_wrps;
	list<ProcWrapper> proc_wrps;
	list<ProcWrapper>::iterator wrp_it;

	list<CommMinNode*> nodes;
	list<CommMinNode*>::iterator node_it;

	//get rid of any old data
	global_graph.clean();
	task_lists.clear();

	//reset all processors
	for(list<Processor *>::iterator proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
	{
		Processor* proc_ptr = *proc_it;

		proc_ptr->Clear_Tasks();							//clear all tasks on the processor
		proc_ptr->Property()->Reset_Available_Space();		//give the processor its entire amount of space available
	}

	PopulateOriginalGraph(&global_graph);
	if((rval = DetermineProcCommMinBounds(proc_wrps)) != 0)
		return rval;	//not all tasks would be able to be mapped to a processor

	//insert processor wrappers into ms_proc_wrps to keep them in order of their ideal bound
	for(wrp_it = proc_wrps.begin(); wrp_it != proc_wrps.end(); ++wrp_it)
		ms_proc_wrps.insert(*wrp_it);

	nodes= global_graph.GetNodes();
	TaskList t_list;

	//insert the nodes' lists into the task)lists to keep them in order of their total utilization
	for(node_it = nodes.begin(); node_it != nodes.end(); ++node_it)
	{
		t_list.tasks = (*node_it)->GetTasks();
		t_list.utilization = (*node_it)->GetTotalUtilization();
		task_lists.insert(t_list);
	}

	//choose which algorithm to run
	if(algo1)
	{
		if(rval = DoAlgo1(ms_proc_wrps))
			return rval;
	}
	else if(algo2)
	{
		if(rval = DoAlgo2(ms_proc_wrps))
			return rval;
	}
	else
		return -1;

	GraphPartition partition;
	partition.SetGroupSize(sys.Get_Number_Of_Procs());

	AssignProcs(ms_proc_wrps);	//assign procs. to each of the tasks that are in the nodes in global_graph
	form_partition(&partition);	//form a partition from the processor assignments
	partition.cost = findPartitionCost(&partition, &global_graph);	//find the cost of this partition

	//print the results
	print_CommMin_Mapping(partition.cost);

	return 0;
}

double DoAlgo1(multiset<ProcWrapper, ProcWrpGreaterThan>& proc_wrps)
{
	multimap<CommMinNode*, CommMinNode*> checked;
	multimap<CommMinNode*, CommMinNode*>::iterator checked_it; 

	list<CommMinNode*>::iterator node_it;
	list<CommMinNode*> nodes = global_graph.GetNodes();

	list<CommMinEdge*>::iterator edge_it;
	list<CommMinEdge*> edges;

	multiset<CommMinNode*, NodeEdgeCostGreaterThan> ms_nodes;
	multiset<CommMinNode*, NodeEdgeCostGreaterThan>::iterator ms_node_it;
	multiset<CommMinNode*, NodeEdgeCostGreaterThan>::iterator ms_node_it1;
	multiset<CommMinNode*, NodeEdgeCostGreaterThan>::iterator ms_node_it2;

	multiset<CommMinEdge*, EdgeCostGreaterThan> ms_edges;
	multiset<CommMinEdge*, EdgeCostGreaterThan>::iterator ms_edge_it;

	//sort nodes by the sum of the cost of each nodes' edges and resize checked
	for(node_it = nodes.begin(); node_it != nodes.end(); ++node_it)
		ms_nodes.insert(*node_it);

	while(global_graph.GetNodes().size() > proc_wrps.size())
	{	
		best_valid_merge.node1 = NULL;
		best_valid_merge.node2 = NULL;
		best_valid_merge.extra_util = 0;

		checked.clear();

		//go through each node starting with the node with the greatest total edge cost
		for(ms_node_it = ms_nodes.begin(); ms_node_it != ms_nodes.end(); ++ms_node_it)
		{
			edges = (*ms_node_it)->GetEdges();
			ms_edges.clear();

			//sort edges by their edge cost
			for(edge_it = edges.begin(); edge_it != edges.end(); ++edge_it)
				ms_edges.insert(*edge_it);

			//go through each edge that is connected to this node starting with the edge with the highest cost
			for(ms_edge_it = ms_edges.begin(); ms_edge_it != ms_edges.end(); ++ms_edge_it)
			{
				CommMinNode* node1 = *ms_node_it;
				CommMinNode* node2 = (*ms_edge_it)->GetOtherNode(*ms_node_it);

				for(checked_it = checked.lower_bound(node1); checked_it != checked.upper_bound(node1); ++checked_it)
					if(checked_it->second == node2) break;
					
				if(checked_it != checked.upper_bound(node1))
					continue;

				checked.insert(make_pair(node1, node2));
				checked.insert(make_pair(node2, node1));

				//check to see if these nodes can be merged - if so merge them
				if(CheckAndMerge(node1, node2, proc_wrps))
				{
					//update ms_nodes to remove merged nodes and to keep nodes in order

					//could not get ms_nodes.find(node2) to work so used this search loop
					for(ms_node_it2 = ms_nodes.begin(); ms_node_it2 != ms_nodes.end(); ++ms_node_it2)
						if(*ms_node_it2 == node2)
						{
							ms_nodes.erase(ms_node_it2);
							break;
						}

					ms_nodes.erase(ms_node_it);	//remove node and reinsert it to keep nodes in order
					ms_nodes.insert(node1);
					break;
				}
			}

			if(ms_edge_it != ms_edges.end())	//nodes were merged
				break;
		}

		ShowNodes();

		if(ms_node_it == ms_nodes.end())
		{
			ShowNodes();

			CommMinNode* node1;
			CommMinNode* node2;
			double extra_util;

			//a valid merge was found that caused some processor to be above its ideal bound
			if(best_valid_merge.node1)
			{
				node1 = best_valid_merge.node1;
				node2 = best_valid_merge.node2;
				extra_util = best_valid_merge.extra_util;
			}
			else
			{
				PossibleMerge merged_nodes;
				if(MergeUnconnected(merged_nodes, ms_nodes, checked, proc_wrps))
				{
					node1 = merged_nodes.node1;
					node2 = merged_nodes.node2;
					extra_util = merged_nodes.extra_util;
				}
				else
					return -1;	//no solution was found
			}

			if(extra_util != -1)
				MergeAndUpdate(node1, node2);

			//could not get ms_nodes.find(node2) to work so used this search loop
			ms_node_it1 = ms_nodes.end();
			ms_node_it2 = ms_nodes.end();

			for(ms_node_it = ms_nodes.begin(); ms_node_it != ms_nodes.end(); ++ms_node_it)
			{
				if(*ms_node_it == node1)
				{
					ms_node_it1 = ms_node_it;
					if(ms_node_it2 != ms_nodes.end())
						break;
				}
				if(*ms_node_it == node2)
				{
					ms_node_it2 = ms_node_it;
					if(ms_node_it1 != ms_nodes.end())
						break;
				}
			}

			ms_nodes.erase(ms_node_it1);			//remove and reinsert node1 to keep nodes in order
			ms_nodes.erase(ms_node_it2);
			ms_nodes.insert(node1);
		}
	}

	ShowNodes();
	return 0;
}

double DoAlgo2(multiset<ProcWrapper, ProcWrpGreaterThan>& proc_wrps)
{
	multimap<CommMinNode*, CommMinNode*> checked;
	multimap<CommMinNode*, CommMinNode*>::iterator checked_it; 

	multiset<CommMinEdge*, EdgeCostGreaterThan> ms_edges;
	multiset<CommMinEdge*, EdgeCostGreaterThan>::iterator ms_edge_it;

	list<CommMinNode*> nodes;
	list<CommMinNode*>::iterator node_it;
	multiset<CommMinNode*, NodeEdgeCostGreaterThan> ms_nodes;

	list<CommMinEdge*>::iterator edge_it;
	list<CommMinEdge*> edges;

	while(global_graph.GetNodes().size() > proc_wrps.size())
	{
		best_valid_merge.node1 = NULL;
		best_valid_merge.node2 = NULL;
		best_valid_merge.extra_util = 0;

		checked.clear();

		edges = global_graph.GetEdges();
		ms_edges.clear();

		//sort edges by their cost
		for(edge_it = edges.begin(); edge_it != edges.end(); ++edge_it)
			ms_edges.insert(*edge_it);

		//go through each edge starting with the one with the highest cost
		for(ms_edge_it = ms_edges.begin(); ms_edge_it != ms_edges.end(); ++ms_edge_it)
		{
			CommMinNode* node1 = (*ms_edge_it)->GetFirstConnectedNode();
			CommMinNode* node2 = (*ms_edge_it)->GetSecondConnectedNode();

			for(checked_it = checked.lower_bound(node1); checked_it != checked.upper_bound(node1); ++checked_it)
				if(checked_it->second == node2) break;
				
			if(checked_it != checked.upper_bound(node1))
				continue;

			checked.insert(make_pair(node1, node2));
			checked.insert(make_pair(node2, node1));

			//see if the two nodes that are part of the edge can be merged - if so merge them
			if(CheckAndMerge(node1, node2, proc_wrps))
				break;
		}

		if(ms_edge_it == ms_edges.end())
		{
			if(best_valid_merge.node1)
				MergeAndUpdate(best_valid_merge.node1, best_valid_merge.node2);
			else
			{
				PossibleMerge merged_nodes;
				nodes = global_graph.GetNodes();

				ms_nodes.clear();
				for(node_it = nodes.begin(); node_it != nodes.end(); ++node_it)
					ms_nodes.insert(*node_it);

				if(MergeUnconnected(merged_nodes, ms_nodes, checked, proc_wrps))
				{
					if(merged_nodes.extra_util != -1)
						MergeAndUpdate(merged_nodes.node1, merged_nodes.node2);
				}
				else
					return -1;	//-1 signifies that the algorithm could not be completed
			}
		}
	}

	return 0;
}

double FindTotalEdgeCost(CommMinNode* node)
{
	double cost = 0;
	list<CommMinEdge*>::iterator edge_it;

	for(edge_it = node->GetEdgeBegin(); edge_it != node->GetEdgeEnd(); ++edge_it)
		cost += (*edge_it)->GetMessageSize();

	return cost;
}

double FindTotalUtilization()
{
	list<CommMinNode*>::iterator node_it;
	list<CommMinNode*> nodes = global_graph.GetNodes();
	double utilization = 0;

	for(node_it = nodes.begin(); node_it != nodes.end(); ++node_it)
		utilization += (*node_it)->GetTotalUtilization();

	return utilization;
}

double DetermineProcCommMinBounds(list<ProcWrapper>& wrappers)
{
	list<ProcWrapper> empty;
	ProcWrapper curr_wrapper;
	list<ProcWrapper>::iterator wrapper_it;
	list<Processor*>::iterator proc_it;
	int non_full_procs = sys.Get_Number_Of_Procs();

	double total_utilization = FindTotalUtilization();
	double util_avg;

	//populate wrappers with a wrapper around each processor
	for(proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
		wrappers.push_back(ProcWrapper(*proc_it));

	//go until there is no more utiliation to distribute
	while(total_utilization && total_utilization > .000001) //second condition takes care of errors when using doubles 
	{
		bool util_changed = false;
		util_avg = total_utilization / non_full_procs;

		for(wrapper_it = wrappers.begin(); wrapper_it != wrappers.end(); ++wrapper_it)
		{
			double ideal_bound = wrapper_it->ideal_bound;
			double true_bound = wrapper_it->proc->Property()->Get_Bound();

			//see if the added util_avg will excede the processor's bound
			if(util_avg + ideal_bound <= true_bound)
			{
				wrapper_it->ideal_bound += util_avg;
				wrapper_it->ideal_space += util_avg;
				total_utilization -= util_avg;
				util_changed = true;
			}
			else if(abs(true_bound - ideal_bound) > .000001)	//if the processors ideal bound does not already equal
			{													//its true bound - errors for doubles accounted for
				double added_bound = true_bound - ideal_bound;	//find the amount the bound can hold
				wrapper_it->ideal_bound += added_bound;
				wrapper_it->ideal_space += added_bound;
				total_utilization -= added_bound;
				non_full_procs--;								//processor is full so there is one fewer non_full_proc
				util_changed = true;
			}
		}

		//if all of the processors are full or there was no utilization change
		if(!non_full_procs || !util_changed)
			return total_utilization;
	}

	return 0;
}

bool CheckAndMerge(CommMinNode* node1, CommMinNode* node2, multiset<ProcWrapper, ProcWrpGreaterThan>& proc_wrps)
{
	list<TaskType*>::iterator task_it;
	multiset<ProcWrapper, ProcWrpGreaterThan>::iterator wrp_it;
	multiset<TaskList, TaskListGreaterThan>::iterator it;
	multiset<TaskList, TaskListGreaterThan>::iterator insert_it;
	multiset<TaskList, TaskListGreaterThan>::iterator it1;
	multiset<TaskList, TaskListGreaterThan>::iterator it2;

	it1 = GetTaskList(node1);				//get the position of the two lists in task_lists
	it2 = GetTaskList(node2);

	TaskList list1 = *it1;					//get the actual lists
	TaskList list2 = *it2;
	TaskList copy_list1 = list1;			//store original version of list1

	task_lists.erase(it1);					//remove lists from task_lists
	task_lists.erase(it2);

	//merge two lists together
	list1.utilization += list2.utilization;
	list1.tasks.insert(list1.tasks.end(), list2.tasks.begin(), list2.tasks.end());

	insert_it = task_lists.insert(list1);	//reinsert list1

	double extra_util = 0;

	it = task_lists.begin();
	for(wrp_it = proc_wrps.begin(); wrp_it != proc_wrps.end(); ++wrp_it)
	{
		double util			= it->utilization;
		double ideal_bound	= wrp_it->ideal_bound;
		double true_bound	= wrp_it->proc->Property()->Get_Bound();

		if(util*.99999 > ideal_bound)	//if the utilization is greater than desired
		{
			if(util > true_bound)		//if the utilization is greater than possible 
			{
				//restore list to its previous state - before list1 and list2 were merged
				task_lists.erase(insert_it);
				task_lists.insert(copy_list1);
				task_lists.insert(list2);
				return false;
			}
			
			extra_util += util - ideal_bound;	//utilization that could excedes the ideal bounds
		}

		++it;
	}

	if(extra_util)	//if the merge causes 1 or more processors to excede their ideal bounds (but not their true bounds)
	{
		//if the extra utilization is smaller then the previous best_valid_merge or 
		//there was no previous best_valid_merge
		if(best_valid_merge.node1 == NULL || extra_util < best_valid_merge.extra_util)
		{
			best_valid_merge.extra_util = extra_util;
			best_valid_merge.node1 = node1;
			best_valid_merge.node2 = node2;
		}

		//restore list to its previous state - before list1 and list2 were merged
		task_lists.erase(insert_it);
		task_lists.insert(copy_list1);
		task_lists.insert(list2);

		return false;
	}

	//everything is ok merge node1 and node2
	node1->Merge(node2);
	global_graph.RemoveNode(node2);
	global_graph.RemoveEdgesFrom(node2);

	return true;
}

double abs(double d)
{
	if(d < 0)
		return -d;
	return d;
}

void ShowNodes()
{
	list<CommMinNode*> nodes = global_graph.GetNodes();
	CString name;
	int group = 0;

	for(list<CommMinNode*>::iterator node_it = nodes.begin(); node_it != nodes.end(); ++node_it)
	{
		for(list<TaskType*>::iterator task_it = (*node_it)->GetTasksBegin(); task_it != (*node_it)->GetTasksEnd(); ++task_it)
			name = (*task_it)->Property()->Get_Name();
		group++;
	}
}

multiset<TaskList, TaskListGreaterThan>::iterator GetTaskList(CommMinNode* node)
{
	multiset<TaskList, TaskListGreaterThan>::iterator list_it;
	TaskType* task;

	//go through each task_list in the list
	for(list_it = task_lists.begin(); list_it != task_lists.end(); ++list_it)
		for(list<TaskType*>::iterator it = list_it->tasks.begin(); it != list_it->tasks.end(); ++it)
		{
			task = *it;
			if(task == *(node->GetTasksBegin())) //if task is equal to the first task in the node
				return list_it;
		}

	return task_lists.end();
}

void MergeAndUpdate(CommMinNode* node1, CommMinNode* node2)
{
	multiset<TaskList, TaskListGreaterThan>::iterator it;
	multiset<TaskList, TaskListGreaterThan>::iterator it1;
	multiset<TaskList, TaskListGreaterThan>::iterator it2;

	it1 = GetTaskList(node1);
	it2 = GetTaskList(node2);

	TaskList list1 = *it1;
	TaskList list2 = *it2;

	task_lists.erase(it1);
	task_lists.erase(it2);

	//merge the two lists
	list1.utilization += list2.utilization;
	list1.tasks.insert(list1.tasks.end(), list2.tasks.begin(), list2.tasks.end());

	task_lists.insert(list1);	//reinsert list1 to keep in order

	//merge actual nodes
	node1->Merge(node2);
	global_graph.RemoveNode(node2);
	global_graph.RemoveEdgesFrom(node2);
}

void AssignProcs(multiset<ProcWrapper, ProcWrpGreaterThan>& proc_wrps)
{
	multiset<ProcWrapper, ProcWrpGreaterThan>::iterator wrp_it;
	multiset<TaskList, TaskListGreaterThan>::iterator list_it;
	list<TaskType*>::iterator task_it;

	//assign list with the greatest utilization to the processor with the greatest ideal bound, etc.
	list_it = task_lists.begin();
	for(wrp_it = proc_wrps.begin(); wrp_it != proc_wrps.end(); ++wrp_it)
	{
		if(list_it == task_lists.end())
			break;

		Processor* proc = wrp_it->proc;

		for(task_it = list_it->tasks.begin(); task_it != list_it->tasks.end(); ++task_it)
		{
			double util = (*task_it)->Property()->Get_Exec_Time() / (*task_it)->Property()->Get_Period();
			proc->Property()->Allocate_Space(util);
			proc->Add_Task(*task_it);

			(*task_it)->Property()->Set_Proc(proc);
		}
		
		++list_it;
	}
}
bool MergeUnconnected(PossibleMerge& merged_nodes, multiset<CommMinNode*, NodeEdgeCostGreaterThan> nodes, multimap<CommMinNode*, CommMinNode*> checked, multiset<ProcWrapper, ProcWrpGreaterThan>& proc_wrps)
{
	multiset<CommMinNode*, NodeEdgeCostGreaterThan>::reverse_iterator node_it1;
	multiset<CommMinNode*, NodeEdgeCostGreaterThan>::reverse_iterator node_it2;
	multimap<CommMinNode*, CommMinNode*>::iterator checked_it;

	best_valid_merge.node1 = NULL;
	best_valid_merge.node2 = NULL;
	best_valid_merge.extra_util = 0;

	CommMinNode* node1;
	CommMinNode* node2;

	for(node_it1 = nodes.rbegin(); node_it1 != nodes.rend(); ++node_it1)
	{
		for(node_it2 = node_it1; node_it2 != nodes.rend(); ++node_it2)
		{
			if(node_it1 == node_it2)
				continue;

			ShowNodes();

			node1 = *node_it1;
			node2 = *node_it2;

			for(checked_it = checked.lower_bound(node1); checked_it != checked.upper_bound(node1); ++checked_it)
				if(checked_it->second == node2) break;
					
			if(checked_it != checked.upper_bound(node1))
				continue;

			checked.insert(make_pair(node1, node2));
			checked.insert(make_pair(node2, node1));

			if(CheckAndMerge(node1, node2, proc_wrps))
			{
				merged_nodes.node1 = node1;
				merged_nodes.node2 = node2;
				merged_nodes.extra_util = -1;
				return true;
			}
		}
	}

	if(best_valid_merge.node1)
	{
		merged_nodes.node1 = best_valid_merge.node1;
		merged_nodes.node2 = best_valid_merge.node2;
		merged_nodes.extra_util = best_valid_merge.extra_util;
		return true;
	}

	return false;
}
