#include "comm_min_graph.h"

//global includes
#include <map>

//local includes
#include "task_connection.h"
#include "processor_network.h"
#include "comm_min-first_fit.h"
#include "graph.h"
#include "system.h"

//externals
extern Graph grp;
extern System sys;

using namespace std;

/****************** CommMinNode Definitions *********************/

CommMinNode::CommMinNode(TaskType* task) : proc(NULL)
{
	added_utilization = 0;
	tasks.push_back(task);
	total_utilization = task->Property()->Get_Exec_Time() / task->Property()->Get_Period();
}

CommMinNode::CommMinNode(CommMinNode* node) : proc(NULL)
{
	added_utilization = 0;
	tasks = node->GetTasks();
	total_utilization = node->GetTotalUtilization();
	proc = node->GetAssignedProc();
}

CommMinNode::~CommMinNode()
{
}

bool CommMinNode::operator == (CommMinNode n)
{
	//make sure it has all of the same tasks
	for(list<TaskType*>::iterator i_it = tasks.begin(); i_it != tasks.end(); ++i_it)
	{
		for(list<TaskType*>::iterator j_it = n.GetTasksBegin(); j_it != n.GetTasksEnd(); ++j_it)
			if(*j_it == *i_it) break;

		if(j_it == n.GetTasksEnd())
			return false;
	}

	return true;
}

bool CommMinNode::Merge(CommMinNode* node)
{
	list<CommMinEdge*>	edge_list = node->GetEdges();
	list<TaskType*>		task_list = node->GetTasks();

	list<CommMinEdge*>::iterator edge_it;

	//check to see if the nodes are connected
	for(edge_it = edge_list.begin(); edge_it != edge_list.end(); ++edge_it)
	{
		if((*edge_it)->GetFirstConnectedNode() == this || (*edge_it)->GetSecondConnectedNode() == this)
			break;
	}

	//nodes are not connected, cannot merge
	if(edge_it == edges.end())
		return false;

	if(added_utilization)	//there was a quick merge done before - undo it
		UndoQuickMerge();

	//add in the tasks from the node to be merged
	for(list<TaskType*>::iterator task_it = task_list.begin(); task_it != task_list.end(); ++task_it)
		tasks.push_back(*task_it);

	//update the total utilization
	total_utilization += node->GetTotalUtilization();

	//add all edges from the node to be merged or add values to existing edges
	for(edge_it = edge_list.begin(); edge_it != edge_list.end(); ++edge_it)
	{
		CommMinEdge* edge_ptr = *edge_it;
		CommMinNode* check_node;

		//find node that is connected to the node to be merged
		if(edge_ptr->GetFirstConnectedNode() == node)
			check_node = edge_ptr->GetSecondConnectedNode();
		else if(edge_ptr->GetSecondConnectedNode() == node)
			check_node = edge_ptr->GetFirstConnectedNode();
		else
			return false;	//node to be merged was not part of this connection - error

		//ignore the connection from the node to be merged to this node
		if(check_node == this)
			continue;

		//check to see if the edge is currently part of this node
		for(list<CommMinEdge*>::iterator cur_edge_it = edges.begin(); cur_edge_it != edges.end(); ++cur_edge_it)
		{
			//edge found add in its weight
			if((*cur_edge_it)->GetFirstConnectedNode() == check_node || 
			   (*cur_edge_it)->GetSecondConnectedNode() == check_node)	
			{
				(*cur_edge_it)->AddMessageSize((*edge_it)->GetMessageSize());
				break;
			}
		}

		//edge was not currently part of this node - add it in
		if(cur_edge_it == edges.end())
		{
			//move the connection from the merged node to this node
			if(edge_ptr->GetFirstConnectedNode() == node)
				edge_ptr->SetFirstConnectedNode(this);
			else if(edge_ptr->GetSecondConnectedNode() == node)
				edge_ptr->SetSecondConnectedNode(this);
	
			edges.push_back(edge_ptr);	//add edge to this node
			node->RemoveEdge(edge_ptr);	//remove edge from old node
		}
	}

	return true;
}

void CommMinNode::QuickMerge(CommMinNode* node)
{
	added_utilization = node->GetTotalUtilization();
	total_utilization += added_utilization;
}

void CommMinNode::UndoQuickMerge()
{
	total_utilization -= added_utilization;
	added_utilization = 0;
}

void CommMinNode::QuickEdgeMerge(CommMinNode* node)
{
	list<CommMinEdge*>::iterator edge_it;
	CommMinEdge* edge;

	if(node == this)
		return;

	for(edge_it = node->GetEdgeBegin(); edge_it != node->GetEdgeEnd(); ++edge_it)
	{
		edge = *edge_it;

		if(edge->GetFirstConnectedNode() == node)
			edge->SetFirstConnectedNode(this);
		if(edge->GetSecondConnectedNode() == node)
			edge->SetSecondConnectedNode(this);

		this->AddEdge(edge);
	}
}

void CommMinNode::QuickNodeMerge(CommMinNode* node)
{
	list<TaskType*>::iterator task_it;

	for(task_it = node->GetTasksBegin(); task_it != node->GetTasksEnd(); ++task_it)
		this->AddTask(*task_it);
}

void CommMinNode::RemoveEdge(CommMinEdge* edge)
{
	for(list<CommMinEdge*>::iterator it = edges.begin(); it != edges.end(); ++it)
		if(edge == *it)
		{
			edges.erase(it);
			return;
		}
}

CommMinEdge* CommMinNode::GetEdgeConnecting(CommMinNode* node)
{
	for(list<CommMinEdge*>::iterator it = edges.begin(); it != edges.end(); ++it)
	{
		CommMinEdge* edge = *it;

		if(edge->GetOtherNode(this) == node)
			return edge;
	}

	return NULL;
}

/*______________________________________________________________*/


/****************** CommMinEdge Definitions *********************/

CommMinEdge::CommMinEdge(double bytes_per_second, CommMinNode* first_node, CommMinNode* second_node)
{
	 bps   = bytes_per_second;
	 node1 = first_node;
	 node2 = second_node; 
}

CommMinNode* CommMinEdge::GetOtherNode(CommMinNode* node)
{
	if(node1 == node)
		return node2;
	if(node2 == node)
		return node1;
	return NULL;
}

/*______________________________________________________________*/


/******************* CommMinGraph Definitions *******************/

CommMinGraph::CommMinGraph(CommMinGraph* graph)
{
	list<CommMinEdge*> copy_edges = graph->GetEdges();
	list<CommMinNode*> copy_nodes = graph->GetNodes();

	for(list<CommMinNode*>::iterator node_it = copy_nodes.begin(); node_it != copy_nodes.end(); ++node_it)
		nodes.push_back(new CommMinNode(*node_it));

	//add in all of the edges
	for(node_it = copy_nodes.begin(); node_it != copy_nodes.end(); ++node_it)
	{
		//add in connections that are coming out of this connection
		for(list<CommMinEdge*>::iterator edge_it = (*node_it)->GetEdgeBegin();
		edge_it != (*node_it)->GetEdgeEnd(); ++edge_it)
		{
			CommMinNode* copyNode1 = (*edge_it)->GetFirstConnectedNode();
			CommMinNode* copyNode2 = (*edge_it)->GetSecondConnectedNode();

			//get a task (any task) from the nodes
			TaskType* task1 = copyNode1->GetTasks().front();
			TaskType* task2 = copyNode2->GetTasks().front();

			//find the new nodes that have these tasks 
			CommMinNode* node1 = this->FindNodeFromTask(task1);
			CommMinNode* node2 = this->FindNodeFromTask(task2);

			if(already_connected(node1, node2))
				continue;

			double bps = (*edge_it)->GetMessageSize();

			CommMinEdge* new_edge = new CommMinEdge(bps, node1, node2);

			//add the new edge to its nodes and to the graph
			node1->AddEdge(new_edge);
			node2->AddEdge(new_edge);
			edges.push_back(new_edge);
		}
	}
}

void CommMinGraph::clean()
{
	//delete all of the nodes
	for(list<CommMinNode*>::iterator node_it = nodes.begin(); node_it != nodes.end(); ++node_it)
		delete *node_it;
	nodes.clear();

	//delete all of the edges
	for(list<CommMinEdge*>::iterator edge_it = edges.begin(); edge_it != edges.end(); ++edge_it)
		delete *edge_it;
	edges.clear();
}

CommMinNode* CommMinGraph::FindNodeFromTask(TaskType* task)
{
	//go through each node in the graph
	for(list<CommMinNode*>::iterator node_it = nodes.begin(); node_it != nodes.end(); ++node_it)
	{
		list<TaskType*> tasks = (*node_it)->GetTasks();

		//go through all of the tasks that are part of the node
		for(list<TaskType*>::iterator task_it = tasks.begin(); task_it != tasks.end(); ++task_it)
			if(*task_it == task) return *node_it;	//found the node
	}

	return NULL;
}

void CommMinGraph::RemoveEdgesFrom(CommMinNode* t)
{
	for(list<CommMinEdge*>::iterator it = t->GetEdgeBegin(); it != t->GetEdgeEnd(); ++it)
	{
		CommMinEdge* edge = *it;

		//remove the edge from the node that is not part of t
		if(edge->GetFirstConnectedNode() == t)
			edge->GetSecondConnectedNode()->RemoveEdge(edge);
		else if(edge->GetSecondConnectedNode() == t)
			edge->GetFirstConnectedNode()->RemoveEdge(edge);
		else
			continue;

		//remove the edge from the graph
		RemoveEdge(edge);

		delete edge;
	}
}

void CommMinGraph::RemoveNode(CommMinNode* node)
{
	for(list<CommMinNode*>::iterator it = nodes.begin(); it != nodes.end(); ++it)
		if(*it == node)
		{
			nodes.erase(it);
			return;
		}
}

void CommMinGraph::RemoveEdge(CommMinEdge* edge)
{
	for(list<CommMinEdge*>::iterator it = edges.begin(); it != edges.end(); ++it)
		if(*it == edge)
		{
			edges.erase(it);
			return;
		}
}

void CommMinGraph::ReAtach(CommMinNode* node, list<attachment> attachments)
{
	nodes.push_back(node);
	list<CommMinNode*> attached_nodes;

	for(list<attachment>::iterator it = attachments.begin(); it != attachments.end(); ++it)
	{
		CommMinNode* attach_node = this->FindNodeFromTask(it->task);

		//make sure this connection isn't already taken care of 
		if(onList(attach_node, attached_nodes))
			continue;
		else
			attached_nodes.push_back(attach_node);

		CommMinEdge* edge = new CommMinEdge(it->bytes_per_second, node, attach_node);

		//add the new edge to the nodes and this graph
		node->AddEdge(edge);
		attach_node->AddEdge(edge);
		this->AddEdge(edge);
	}
}

/*______________________________________________________________*/


/******************* Graph Partition Defintions *****************/

GraphPartition::GraphPartition(const GraphPartition& g)
{
	cost = g.cost;
	num_populated = g.num_populated;

	SetGroupSize(g.group.size());

	//copy the groups and the procs
	for(int i = 0; i < g.group.size(); ++i)
	{
		group[i] = g.group[i];
		procs[i] = g.procs[i];
	}

}

GraphPartition& GraphPartition::operator = (const GraphPartition& g)
{
	if(this != &g)	//make sure g isn't this graph 
	{
		cost = g.cost;
		
		SetGroupSize(g.group.size());

		for(int i = 0; i < g.group.size(); ++i)
		{
			group[i] = g.group[i];
			procs[i] = g.procs[i];
		}
	}

	return *this;
}

void GraphPartition::AddTasks(int group_num, CommMinNode* node)
{
	for(list<TaskType*>::iterator it = node->GetTasksBegin(); it != node->GetTasksEnd(); ++it)
		group[group_num].push_back(*it);
}

double GraphPartition::GetTotalUtilization(int group_num)
{
	double utilization = 0;

	for(list<TaskType*>::iterator it = group[group_num].begin(); it != group[group_num].end(); ++it)
		utilization += (*it)->Property()->Get_Exec_Time() / (*it)->Property()->Get_Period();

	return utilization;
}

void GraphPartition::SetGroupSize(int size)
{
	int cur_size = group.size();

	group.resize(size); 
	procs.resize(size);

	//set all the new processors to NULL
	for(int i = cur_size; i < size; ++i)
		procs[i] = NULL;
}
/*______________________________________________________________*/

bool already_connected(CommMinNode* node1, CommMinNode* node2)
{
	for(list<CommMinEdge*>::iterator it = node1->GetEdgeBegin(); it != node1->GetEdgeEnd(); ++it)
	{
		CommMinNode* check_node;

		if((*it)->GetFirstConnectedNode() == node1)
			check_node = (*it)->GetSecondConnectedNode();
		else
			check_node = (*it)->GetFirstConnectedNode();

		if(check_node == node2)
			return true;
	}

	return false;
}

void PopulateGraph(CommMinGraph* graph)
{
	graph->clean();		//clear out any existing stuff

	//add in all of the nodes - needs to be done before adding edges
	for(list<TaskType*>::iterator task_it = grp.Node_Begin(); task_it != grp.Node_End(); ++task_it)
	{
		TaskType* task_ptr = *task_it;
		CommMinNode* new_task = new CommMinNode(task_ptr);
		graph->AddNode(new_task);
	}

	//add in all of the edges
	for(task_it = grp.Node_Begin(); task_it != grp.Node_End(); ++task_it)
	{
		//add in connections that are coming out of this connection
		for(list<ConnectionType*>::iterator conn_it = (*task_it)->Get_Out_Data_Edges_Begin(); 
		conn_it != (*task_it)->Get_Out_Data_Edges_End(); ++conn_it)
		{
			TaskType* from = (*conn_it)->Get_From();
			TaskType* to = (*conn_it)->Get_To();
			double freq = (*conn_it)->Property()->Get_Frequency();

			//get the node that the tasks are part of 
			CommMinNode* node1 = graph->FindNodeFromTask(from);
			CommMinNode* node2 = graph->FindNodeFromTask(to);

			double bps = (*conn_it)->Property()->Get_Frequency() * (*conn_it)->Property()->Get_Message_Size();

			CommMinEdge* new_edge = new CommMinEdge(bps, node1, node2);
			new_edge->SetFrequency(freq);

			//add the new edge to its nodes and to the graph
			node1->AddEdge(new_edge);
			node2->AddEdge(new_edge);
			graph->AddEdge(new_edge);
		}
	}
}

void MergeNodesWithProcs(CommMinGraph* graph)
{
	list<CommMinNode*> nodes;
	list<CommMinNode*>::iterator node_it;
	list<Processor*>::iterator proc_it;
	CommMinNode* merged_node;

	map<Processor*, list<CommMinNode*> > proc_to_nodes;
	Processor* proc;

	//assign all of the nodes to a processor if they have one assigned to them
	nodes = graph->GetNodes();
	for(node_it = nodes.begin(); node_it != nodes.end(); ++node_it)
	{
		CommMinNode* curr_node = *node_it;
		TaskType* first_task = *curr_node->GetTasksBegin();

		if(proc = first_task->Property()->Get_Alloc_Proc())
			proc_to_nodes[proc].push_back(curr_node);
	}

	//go through each processor
	for(proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
	{
		proc = *proc_it;
		merged_node = NULL;

		//merge all of the nodes on this processor
		for(node_it = proc_to_nodes[proc].begin(); node_it != proc_to_nodes[proc].end();)
		{
			CommMinNode* curr_node = *node_it;

			if(merged_node)					//a previous node on this processor was found
			{
				//merge the two nodes
				merged_node->Merge(curr_node);		
				graph->RemoveNode(curr_node);
				graph->RemoveEdgesFrom(curr_node);
				node_it = proc_to_nodes[proc].erase(node_it); //erase this node from the list and get the next iterator position
			}
			else
			{
				merged_node = curr_node;	//first node for this processor found
				++node_it;
			}
		}

		//if there was at least one node on this procssor set that node's processor
		if(merged_node)
			merged_node->AssignProc(proc);
	}
}

void GetNodesWithNoProcs(list<CommMinNode*>& nodes_wo_procs, CommMinGraph* graph)
{
	list<CommMinNode*> nodes;
	list<CommMinNode*>::iterator node_it;

	nodes = graph->GetNodes();
	for(node_it = nodes.begin(); node_it != nodes.end(); ++node_it)
		if(!(*node_it)->GetAssignedProc())
			nodes_wo_procs.push_back(*node_it);
}

void GetNodesWithProcs(std::list<CommMinNode*>& nodes_with_procs, CommMinGraph* graph)
{
	list<CommMinNode*> nodes;
	list<CommMinNode*>::iterator node_it;

	nodes = graph->GetNodes();
	for(node_it = nodes.begin(); node_it != nodes.end(); ++node_it)
		if((*node_it)->GetAssignedProc())
			nodes_with_procs.push_back(*node_it);
}

void SetRefsInEdgesTo(CommMinNode* original_node, CommMinNode* new_node)
{
	list<CommMinEdge*>::iterator edge_it;
	CommMinEdge* edge;

	for(edge_it = original_node->GetEdgeBegin(); edge_it != original_node->GetEdgeEnd(); ++edge_it)
	{
		edge = *edge_it;
		if(edge->GetFirstConnectedNode() == original_node)
			edge->SetFirstConnectedNode(new_node);
		else if(edge->GetSecondConnectedNode() == original_node)
			edge->SetSecondConnectedNode(new_node);
	}
}