#ifndef COMM_MIN_GRAPH_H
#define COMM_MIN_GRAPH_H

#include <list>
#include <vector>

class ConnectionType;
class TaskType;
class Processor;
class CommMinEdge;
class Processor;

/* A simple structure so a node can easily store the values that are make up an edge*/
struct attachment
{
	TaskType* task;
	double bytes_per_second;
};

/* CommMinNode - a node used for the communication minimization algorithm.
   Has a list of tasks that make up the node and a list of CommMinEdges that
   connects the node to other nodes*/
class CommMinNode
{
public:
	CommMinNode() : proc(NULL), total_utilization(0), added_utilization(0) {}
	CommMinNode(TaskType* task);
	CommMinNode(CommMinNode* node);
	~CommMinNode();
	bool Merge(CommMinNode* node);	//merge a node with this node - this means that
									//this node will take all the edges that come
									//from the merged node and the edges between this
									//node and the given node are removed

	void QuickMerge(CommMinNode* node);	//Only adds the nodes utilization to the total utilization
	void UndoQuickMerge();				//Undos the last quick merge

	void QuickEdgeMerge(CommMinNode* node);	//adds all of the given nodes edges to this node
											//and replaces the node referenced in the edge
											//with this node
	void QuickNodeMerge(CommMinNode* node);	//add all tasks in the current node to
											//this node

	bool operator == (CommMinNode n);			 

	inline void AddEdge(CommMinEdge* edge)					{edges.push_back(edge);}
	inline void AddTask(TaskType* task)						{tasks.push_back(task);}

	void RemoveEdge(CommMinEdge* edge);
	inline std::list<CommMinEdge*>	GetEdges()				{return edges;}

	inline std::list<CommMinEdge*>::iterator GetEdgeBegin() {return edges.begin();}
	inline std::list<CommMinEdge*>::iterator GetEdgeEnd()	{return edges.end();}

	inline std::list<TaskType*>::iterator GetTasksBegin()	{return tasks.begin();}
	inline std::list<TaskType*>::iterator GetTasksEnd()		{return tasks.end();}

	inline std::list<TaskType*>	GetTasks()					{return tasks;}
	inline void AssignProc(Processor* processor)			{proc = processor;}

	inline double GetTotalUtilization()						{return total_utilization;}
	inline Processor* GetAssignedProc()						{return proc;}

	CommMinEdge* GetEdgeConnecting(CommMinNode* node);
private:

	std::list<CommMinEdge*>	edges;
	std::list<TaskType*>	tasks;
	double					total_utilization;
	double					added_utilization;
	Processor*				proc;
};

/* CommMinEdge - an undirected edge used in the communication minimization algorithm.
   Has the two nodes that the edge connectes and the 'cost' of the edge which is the
   edges' bytes per second*/
class CommMinEdge
{
public:
	CommMinEdge(double bytes_per_second, CommMinNode* first_node, CommMinNode* second_node);

	inline void AddMessageSize(double bytes_per_second)		{bps += bytes_per_second;}
	inline double GetMessageSize()							{return bps;}
	inline void SetFirstConnectedNode(CommMinNode* node)	{node1 = node;}
	inline void SetSecondConnectedNode(CommMinNode* node)	{node2 = node;}
	inline CommMinNode* GetFirstConnectedNode()				{return node1;}
	inline CommMinNode* GetSecondConnectedNode()			{return node2;}
	inline double GetFrequency()							{return freq;}
	inline void SetFrequency(double frequency)				{freq = frequency;}
	CommMinNode*		GetOtherNode(CommMinNode* node);	//returns the node in the connection that is 
															//not the given node

private:
	CommMinEdge();

	CommMinNode* node1;
	CommMinNode* node2;
	double bps;
	double freq;
};

/* CommMinGraph - a graph used for the communication minimization algorithm.  Has a 
   list of edges and nodes that makes up the graph*/
class CommMinGraph
{
public:
	CommMinGraph() {}
	CommMinGraph(CommMinGraph* graph);			//copy the given graph into this graph
	~CommMinGraph()								{clean();}

	void clean();
	inline std::list<CommMinNode*> GetNodes()	{return nodes;}
	inline std::list<CommMinEdge*> GetEdges()	{return edges;}

	inline std::list<CommMinNode*>::iterator Nodes_Begin()	{return nodes.begin();}
	inline std::list<CommMinNode*>::iterator Nodes_End()	{return nodes.end();}

	inline std::list<CommMinEdge*>::iterator Edges_Begin()	{return edges.begin();}
	inline std::list<CommMinEdge*>::iterator Edges_End()	{return edges.end();}

	inline void AddNode(CommMinNode* node)		{nodes.push_back(node);}
	void RemoveNode(CommMinNode* node);

	inline void AddEdge(CommMinEdge* edge)		{edges.push_back(edge);}
	void RemoveEdge(CommMinEdge* edge);

	CommMinNode* FindNodeFromTask(TaskType* task);	//find the node that a given task is in
	void		 RemoveEdgesFrom(CommMinNode* t);	//remove all the edges from the graph that are part of the given CommMinNode
	void		 ReAtach(CommMinNode* node, std::list<attachment> attachments); //reatch the given node using the list of attachments
private:
	std::list<CommMinNode*> nodes;
	std::list<CommMinEdge*> edges;
};

/* GraphPartition - a division of a CommMinGraph into groups.  Has a cost which is the
   total edge cost for the division of all the groups*/
struct GraphPartition
{
	GraphPartition() {}
	GraphPartition(const GraphPartition& g);
	GraphPartition& operator = (const GraphPartition& g);
	~GraphPartition() {}

	void AddTasks(int group_num, CommMinNode* node); //add all the tasks from the given node to the given group number 

	double GetTotalUtilization(int group_num);	//get the total utilization from all of the nodes in the given group number
	void SetGroupSize(int size);

	std::vector<std::list<TaskType*> >	group;
	std::vector<Processor*>				procs;
	double								cost;
	int									num_populated;
};

/*** Functions for working with CommMinGraphs ***/

//function to test if node1 is already connected to node2 by a CommMinEdge
bool already_connected(CommMinNode* node1, CommMinNode* node2);

//populate the given graph with the nodes and edges from the global Graph grp 
void PopulateGraph(CommMinGraph* graph);

//merge all of the nodes that have tasks that are already assigned to a processor
//and set the node's processor variable
void MergeNodesWithProcs(CommMinGraph* graph);

//populate nodes_wo_procs with all of the nodes in the given graph that do not have
//a processor assigned to them
void GetNodesWithNoProcs(std::list<CommMinNode*>& nodes_wo_procs, CommMinGraph* graph);

//populate nodes_with_procs with all of the nodes in the given graph that do have
//a processor assigned to them
void GetNodesWithProcs(std::list<CommMinNode*>& nodes_with_procs, CommMinGraph* graph);

//Set all of the references to the original node in the edges to be the new_node
void SetRefsInEdgesTo(CommMinNode* original_node, CommMinNode* new_node);
#endif