#ifndef FIRSTFIT_H
#define FIRSTFIT_H

/* First Fit Algorithm
 * -------------------
 *
 * Given a list of different processors and a list of tasks to be completed
 * by these processors find the minimum number of processors needed to complete
 * all of the tasks.  This is achieved by first finding the utilization of each
 * task by dividing the execution time by the period of the task.  All of these tasks
 * are then sorted by this utilization.  The processors are then sorted by their bounds.  
 * For each task the first processor in the list (the one with the highest bound)
 * is checked to determine if the task can be completed on it.  This is determined
 * by checking if the sum of all the utilizations for all of the tasks already 
 * assigned to this processor is less then the processor's bound (this must be less
 * than or equal to 1 and greater than 0).  If the task cannot be put on the first
 * processor then the second processor is tried and so on.  If the task cannot be
 * put on any processor then this is an error and the bounds of the current processors
 * need to be altered or more processors are needed.
 *
 * NOTE: Compiling first_fit.cpp will cause many warnings if you are using MSV C++
 * since the file uses multisets with function objects, which are not handled very well
 * in MSV C++
 */

#include <list>
#include <string>
#include <map>

class TaskType;
class Processor;

//run the first fit algorithm and print the results
double Do_FirstFit_Mapping();

//map the tasks acording to the first fit algorithm
double FirstFit();

//function object to determine if the utility for task t1 is greater than 
//the utility for task t2 (used for sorting)
class task_greater_than
{
public:
	bool operator () (TaskType* t1, TaskType* t2);
};

//function object to determine if the bound for processor p1 is greater than 
//the bound for processor p2 (used for sorting)
class processor_greater_than
{
public:
	bool operator () (Processor* p1, Processor* p2);
};

//same as above but with temp_avail_space
class temp_processor_greater_than
{
public:
	bool operator () (Processor* p1, Processor* p2);
};
#endif